'''
Defines the Level class.
'''

from .constants import *

class Level:
    '''
    Simple data structure for storing the spin (total J), parity (+/-1),
    energy (MeV, excitation), and width (eV) of a level used in an AZURE2
    calculation.

    channel : channel pair (defined in AZURE2)
    radius  : channel radius
    index   : Which spin^{parity} level is this? (There are frequently more than
              one. Consistent with the language, these are zero-based.)
    '''
    def __init__(self, row_str):
        row = row_str.split()
        self.spin = float(row[J_INDEX])
        self.parity = int(row[PI_INDEX])
        self.energy = float(row[ENERGY_INDEX])
        self.energy_fixed = int(row[ENERGY_FIXED_INDEX])
        self.width = float(row[WIDTH_INDEX])
        self.width_fixed = int(int(row[WIDTH_FIXED_INDEX]) or self.width == 0)
        self.channel_radius = float(row[CHANNEL_RADIUS_INDEX])
        self.channel = int(row[CHANNEL_INDEX])
        self.separation_energy = float(row[SEPARATION_ENERGY_INDEX])
        self.include = int(row[LEVEL_INCLUDE_INDEX])

    def print(self):
        '''
        Prints a description of the level.
        '''
        sign = '+' if self.parity > 0 else '-'
        print(f'{self.spin}{sign} | \
{self.energy} MeV | {self.width} eV | channel {self.channel}')



class Parameter:
    '''
    Defines a sampled (or "free") parameter by spin, parity, channel,
    rank, and whether it's an energy or width (kind).
    kind    : "energy" or "width"
              "width" can be the partial width or ANC (depending on how it was
              set up in AZURE2)
    channel : channel pair (defined in AZURE2; consistent with AZURE2,
              these are one-based)
    rank    : Which spin^{parity} level is this? (There are frequently
              more than one. Consistent with AZURE2, these are 
              one-based.)
    '''
    def __init__(self, spin, parity, kind, channel, rank=1, is_anc=False):
        self.spin = spin
        self.parity = parity
        self.kind = kind
        self.channel = int(channel)
        self.rank = rank
        
        jpi_label = '+' if self.parity == 1 else '-'
        subscript = f'{rank:d},{channel:d}'
        superscript = f'({jpi_label:s}{spin:.1f})'
        if self.kind == 'energy':
            self.label = r'$E_{%s}^{%s}$' % (subscript, superscript)
        elif self.kind == 'width':
            if is_anc:
                self.label = r'$C_{%s}^{%s}$' % (subscript, superscript)
            else:
                self.label = r'$\Gamma_{%s}^{%s}$' % (subscript, superscript)


    def string(self):
        '''
        Returns a string that describes the parameter.
        '''
        parity = '+' if self.parity == 1 else '-'
        return f'{self.spin}{parity} {self.kind} (number {self.rank}) in \
particle pair {self.channel}, {self.label}'


    def print(self):
        '''
        Prints a description of the parameters.
        '''
        print(self.string())


class NormFactor:
    '''
    Defines a sampled normalization factor (n_i in the AZURE2 manual).
    '''
    def __init__(self, dataset_index):
        self.index = dataset_index
        self.label = r'$n_{%d}$' % (self.index+1)



'''
Defines classes for interacting with AZURE2.
'''

import os
import shutil
import numpy as np
from . import level
from . import utility
from .parameter import Parameter
from .output import Output
from .data import Data
from .nodata import Test
from .configuration import Config

class AZR:
    '''
    Object that manages the communication between Python and AZURE2.

    Attributes specified at instantiation:
    input_filename          : .azr file
    parameters              : list of Parameter instances (sampled parameters)
    output_filenames        : Which output files (AZUREOut_*.out) are read?
    extrap_filenames        : Which output files (AZUREOut_*.extrap) are read?

    Other attributes (given default values below):
    use_brune               : Bool that indicates the use of the Brune
                              parameterization.
    use_gsl                 : Bool that indicates the use of GSL Coulomb functions.
    ext_par_file            : Filename where parameter values can be read.
    ext_capture_file        : Filename where external capture integral results have
                              been stored.
    ext_capture_file_extrap : Filenme where external capture integral results
                              for segments without data have been stored.
    command                 : Name of AZURE2 binary.
    '''
    def __init__(self, input_filename, parameters=None, output_filenames=None,
                 extrap_filenames=None):
        # Give default values to attributes that are not specified at
        # instantiation. These values must be changed *after* instantiation.
        self.use_brune = True
        self.use_gsl = True
        self.ext_par_file = '\n'
        self.ext_capture_file = '\n'
        self.ext_capture_file_extrap = '\n'
        self.command = 'AZURE2'
        self.root_directory = ''
        self.verbose = True
        
        self.config = Config(input_filename, parameters=parameters)

        '''
        If parameters are not specified, they are inferred from the input file.
        '''
        if parameters is None:
            self.parameters = self.config.parameters.copy()
        else:
            self.parameters = parameters.copy()

        '''
        If output files are not specified, they are inferred from the input file.
        '''
        if output_filenames is None:
            self.output_filenames = self.config.data.output_files
        else:
            self.output_filenames = output_filenames

        '''
        If extrapolation files are not specified, they are inferred from the input file.
        '''
        if extrap_filenames is None:
            self.extrap_filenames = self.config.test.output_files
        else:
            self.extrap_filenames = extrap_filenames


    def predict(self, theta, mod_data=None, dress_up=True, full_output=False):
        '''
        Takes:
            * a point in parameter space, theta.
            * dress_up    : Use Output class.
            * full_output : Return reduced width amplitudes as well.
            * mod_data    : Do any parametes in theta modify the original data?
        Does:
            * creates a random filename ([rand].azr)
            * creates a (similarly) random output directory (output_[rand]/)
            * writes the new Levels to a [rand].azr
            * writes output directory to [rand].azr
            * runs AZURE2 with [rand].azr
            * reads observable from output_[rand]/output_filename
            * deletes [rand].azr
            * deletes output_[rand]/
            * deletes data_[rand]/
        Returns:
            * predicted values and (optionally) reduced width amplitudes.
        '''

        workspace = self.config.generate_workspace(
            theta,
            prepend=self.root_directory,
            mod_data=mod_data
        )
        input_filename, output_dir, data_dir = workspace

        try:
            response = utility.run_AZURE2(input_filename, choice=1,
                use_brune=self.use_brune, ext_par_file=self.ext_par_file,
                ext_capture_file=self.ext_capture_file, use_gsl=self.use_gsl,
                command=self.command)
        except:
            shutil.rmtree(output_dir)
            shutil.rmtree(data_dir)
            os.remove(input_filename)
            if self.verbose:
                print('AZURE2 did not execute properly.')
            raise

        try:
            if dress_up:
                output = [Output(output_dir + '/' + of) for of in
                          self.output_filenames]
            else:
                output = [np.loadtxt(output_dir + '/' + of) for of in
                          self.output_filenames]

            if full_output:
                output = (output, utility.read_rwas_jpi(output_dir))

            shutil.rmtree(output_dir)
            shutil.rmtree(data_dir)
            os.remove(input_filename)

            return output
        except:
            shutil.rmtree(output_dir)
            shutil.rmtree(data_dir)
            os.remove(input_filename)
            if self.verbose:
                print('Output files were not properly read.')
                print('AZURE output:')
                print(response)
            raise


    def extrapolate(self, theta, segment_indices=None, use_brune=None,
                    use_gsl=None, ext_capture_file=None):
        '''
        See predict() documentation.
        '''
        workspace = self.config.generate_workspace_extrap(theta,
            segment_indices=segment_indices)
        input_filename, output_dir, output_files = workspace

        try:
            response = utility.run_AZURE2(input_filename, choice=3,
                use_brune=use_brune if use_brune is not None else self.use_brune,
                use_gsl=use_gsl if use_gsl is not None else self.use_gsl,
                ext_par_file=self.ext_par_file,
                ext_capture_file=(ext_capture_file if ext_capture_file is not
                    None else self.ext_capture_file_extrap),
                command=self.command)
        except:
            shutil.rmtree(output_dir)
            os.remove(input_filename)
            if self.verbose:
                print('AZURE2 did not execute properly.')
            raise

        try:
            output = [np.loadtxt(output_dir + '/' + of) for of in output_files]
            shutil.rmtree(output_dir)
            os.remove(input_filename)
            return output
        except:
            shutil.rmtree(output_dir)
            os.remove(input_filename)
            if self.verbose:
                print('Output files could not be read.')
            raise


    def rwas(self, theta):
        '''
        Returns the reduced width amplitudes (rwas) and their corresponding J^pi
        at the point in parameter space, theta.
        '''
        input_filename, output_dir = utility.random_output_dir_filename()
        new_levels = self.config.generate_levels(theta)
        utility.write_input_file(self.config.input_file_contents, new_levels,
                                 input_filename, output_dir)
        response = utility.run_AZURE2(input_filename, choice=1,
            use_brune=self.use_brune, ext_par_file=self.ext_par_file,
            ext_capture_file=self.ext_capture_file, use_gsl=self.use_gsl,
            command=self.command)

        rwas = utility.read_rwas_jpi(output_dir)

        shutil.rmtree(output_dir)
        os.remove(input_filename)

        return rwas

    
    def ext_capture_integrals(self, use_gsl=False, mod_data=False):
        '''
        Returns the AZURE2 output of external capture integrals.
        '''
        input_filename, output_dir, data_dir = utility.random_workspace()

        if mod_data:
            self.config.update_data_directories(data_dir)

        new_levels = self.config.initial_levels.copy()
        new_levels = [l for sl in new_levels for l in sl]
        utility.write_input_file(self.config.input_file_contents, new_levels,
                                 input_filename, output_dir)
        response = utility.run_AZURE2(input_filename, choice=1,
            use_brune=self.use_brune, ext_par_file=self.ext_par_file,
            ext_capture_file='\n', use_gsl=use_gsl,
            command=self.command)

        ec = utility.read_ext_capture_file(output_dir + '/intEC.dat')

        shutil.rmtree(output_dir)
        shutil.rmtree(data_dir)
        os.remove(input_filename)

        return ec

    
    def update_ext_capture_integrals(self, segment_indices, shifts, use_gsl=False):
        '''
        Takes:
          * a list of indices to identification which data segment is being
            shifted
          * a list of shifts to be applied (in the same order as the indices are
            provided)
        * Adjusts the energies of data segments (identified by index) by the
        provided shifts (MeV, lab).
        * Evaluates the external capture (EC) integrals.
        * Returns the values from the EC file.
        '''
        for (i, shift) in zip(segment_indices, shifts):
            self.config.data.segments[i].shift_energies(shift)

        return self.ext_capture_integrals(use_gsl=use_gsl, mod_data=True)



'''
Classes to hold Data segments as found in .azr files.
'''

import numpy as np
from . import utility
from .parameter import NormFactor
from .constants import *

class Segment:
    '''
    Structure to organize the information contained in a line in the
    <segmentsData> section of an AZURE2 input file.
    '''
    def __init__(self, row, index):
        self.row = row.split()
        self.include = (int(self.row[INCLUDE_INDEX]) == 1)
        self.in_channel = int(self.row[IN_CHANNEL_INDEX])
        self.out_channel = int(self.row[OUT_CHANNEL_INDEX])
        self.norm_factor = float(self.row[NORM_FACTOR_INDEX])
        self.vary_norm_factor = int(self.row[VARY_NORM_FACTOR_INDEX])
        self.index = index

        self.filepath = self.row[FILEPATH_INDEX]
        i = self.filepath.find('/')
        self.filename = self.filepath[i+1:]

        if self.vary_norm_factor:
            self.nf = NormFactor(self.index)
        else:
            self.nf = None
        
        self.values_original = np.loadtxt(self.filepath)
        self.values = np.copy(self.values_original)
        self.n = self.values.shape[0]

        if self.out_channel != -1:
            self.output_filename = f'AZUREOut_aa={self.in_channel}_R={self.out_channel}.out'
        else:
            self.output_filename = f'AZUREOut_aa={self.in_channel}_TOTAL_CAPTURE.out'

   
    def string(self):
        '''
        Returns a string of the text in the segment line.
        '''
        row = self.row.copy()
        # Are these lines...
        row[INCLUDE_INDEX] = '1' if self.include else '0'
        row[IN_CHANNEL_INDEX] = str(self.in_channel)
        row[OUT_CHANNEL_INDEX] = str(self.out_channel)
        row[FILEPATH_INDEX] = str(self.filepath)
        row[NORM_FACTOR_INDEX] = str(self.norm_factor)
        # necessary?
        
        return ' '.join(row)


    def print(self):
        '''
        Prints description of the data segment.
        '''
        print(self.string())


    def update_dir(self, new_dir, values=None):
        '''
        Updates the path directory of the segment.
        If modifications are made to the data, the modified data is written to
        an ephemeral directory so that multiple processes can do so
        simultaneously.
        '''
        filepath = new_dir + '/' + self.filename
        if values is not None:
            np.savetxt(filepath, values)
        else:
            np.savetxt(filepath, self.values)


    def shift_energies(self, shift):
        values = np.copy(self.values_original)
        values[:, 0] += shift
        return values


class Data:
    '''
    Structure to hold all of the data segments in a provided AZURE2 input file.
    '''
    def __init__(self, filename):
        self.contents = utility.read_input_file(filename)
        i = self.contents.index('<segmentsData>')+1
        j = self.contents.index('</segmentsData>')

        # All segments listed in the file.
        self.segments = []
        k = 0
        for row in self.contents[i:j]:
            if row != '':
                row_list = row.split()
                include = (int(row_list[INCLUDE_INDEX]) == 1)
                if include:
                    self.segments.append(Segment(row, k))
                k += 1

        # Indices of segments with varied normalization constants.
        self.norm_segment_indices = []
        for (i, seg) in enumerate(self.segments):
            if seg.include and seg.vary_norm_factor:
                self.norm_segment_indices.append(i)

        # Number of data points for each included segment.
        self.ns = [seg.n for seg in self.segments] 

        # Output files that need to be read.
        self.output_files = []
        for seg in self.segments:
            self.output_files.append(seg.output_filename)
        # Eliminates repeated output files AND SORTS them:
        # (1, 2, 3, ..., TOTAL_CAPTURE)
        self.output_files = list(np.unique(self.output_files))


    def update_all_dir(self, new_dir, contents):
        '''
        Updates all the path directories of the segments.
        '''
        start = contents.index('<segmentsData>')+1
        stop = contents.index('</segmentsData>')

        new_contents = contents.copy()

        for i in range(start, stop):
            row = contents[i].split()
            old_path = row[FILEPATH_INDEX]
            j = old_path.find('/') + 1
            row[FILEPATH_INDEX] = new_dir + '/' + old_path[j:]
            new_contents[i] = ' '.join(row)
        
        for seg in self.segments:
            seg.update_dir(new_dir)

        return new_contents



    def write_segments(self, contents):
        '''
        Writes the segments to contents.
        "contents" is a representation of the .azr file (list of strings)
        This is typically done in preparation for writing a new .azr file.
        '''
        start = contents.index('<segmentsData>')+1
        stop = contents.index('</segmentsData>')

        for segment in self.segments:
            contents[start+segment.index] = segment.string()

        return contents


    def update_norm_factors(self, theta_norm, contents):
        assert len(theta_norm) == len(self.norm_segment_indices), '''
Number of normalization factors does not match the number of data segments
indicating the normalization factor should be varied.
'''
        for (f, i) in zip(theta_norm, self.norm_segment_indices):
            self.segments[i].norm_factor = f

        self.write_segments(contents)
        
        return contents



'''
Class to read and write the contents of an AZURE2 input file.
The purpose is to remove as much of this work from AZR as possible.
'''

from . import utility
from .data import Data
from .nodata import Test
from .parameter import Parameter

class Config:
    def __init__(self, input_filename, parameters=None):
        self.input_filename = input_filename
        self.input_filename = input_filename
        self.input_file_contents = utility.read_input_file(input_filename)
        self.initial_levels = utility.read_levels(input_filename)
        self.data = Data(self.input_filename)
        self.test = Test(self.input_filename)

        if parameters is None:
            self.parameters = []
            jpis = []
            for group in self.initial_levels:
                # grab the J^pi from the first row in the group
                jpi = group[0].spin*group[0].parity
                # add it to the list
                jpis.append(jpi)
                for (i, sublevel) in enumerate(group):
                    if sublevel.include:
                        spin = sublevel.spin
                        parity = sublevel.parity
                        rank = jpis.count(jpi)
                        if i == 0:
                            if not sublevel.energy_fixed:
                                self.parameters.append(
                                    Parameter(
                                        spin, parity, 'energy', i+1, rank=rank
                                    )
                                )
                        if not sublevel.width_fixed:
                            if sublevel.energy < sublevel.separation_energy:
                                self.parameters.append(
                                    Parameter(
                                        spin, parity, 'width', i+1, rank=rank, is_anc=True
                                    )
                                )
                            else:
                                self.parameters.append(
                                    Parameter(
                                        spin, parity, 'width', i+1, rank=rank
                                    )
                                )
        else:
            self.parameters = parameters

        Jpi = [l[0].spin*l[0].parity for l in self.initial_levels]
        self.addresses = []
        for p in self.parameters:
            jpi = p.spin*p.parity
            i = Jpi.index(jpi)
            i += p.rank-1 # convert from one-based count to zero-based index
            j = p.channel-1 # convert from one-based count to zero-based index
            self.addresses.append([i, j, p.kind])

        self.n1 = len(self.parameters)
        self.n2 = len(self.data.norm_segment_indices)
        # number of free parameters
        self.nd = self.n1 + self.n2

        self.labels = []
        for i in range(self.n1):
            self.labels.append(self.parameters[i].label)
        for i in self.data.norm_segment_indices:
            self.labels.append(self.data.segments[i].nf.label)


    def generate_levels(self, theta):
        levels = self.initial_levels.copy()
        for (theta_i, address) in zip(theta, self.addresses):
            i, j, kind = address
            if kind == 'energy':
                '''
                Set the energy for each channel in this level to the prescribed
                energy.
                '''
                for sl in levels[i]:
                    sl.energy = theta_i
            else:
                setattr(levels[i][j], kind, theta_i)
        return [l for sublevel in levels for l in sublevel]


    def get_input_values(self):
        '''
        Returns the values of the sampled parameters in the input file.
        '''
        values = [getattr(self.initial_levels[i][j], kind) for (i, j, kind) in
                  self.addresses]
        for i in self.data.norm_segment_indices:
            values.append(self.data.segments[i].norm_factor)
        return values


    def update_data_directories(self, new_dir, contents):
        '''
        The data needs to be stored in a new location (new_dir), so the input
        has to reflect that. In preparation, the contents of the input file are
        updated here.
        '''
        contents = self.data.update_all_dir(new_dir, contents)
        return self.data.write_segments(contents)


    def generate_workspace(self, theta, prepend='', mod_data=None):
        '''
        Config handles the configuration of the calculation. That includes:
        * mapping theta to the relevant values in the input file
        * setting up the appropriate workspace for AZR to operate in
        '''
        contents = self.input_file_contents.copy()

        new_levels = self.generate_levels(theta[:self.n1])
        contents = self.data.update_norm_factors(theta[self.n1:self.n1+self.n2],
            contents)

        input_filename, output_dir, data_dir = utility.random_workspace(prepend=prepend)


        if mod_data is not None:
            utility.write_input_file(contents, new_levels, input_filename,
                output_dir, data_dir=data_dir)
            self.data.update_all_dir(data_dir, contents)
            for (i, data) in mod_data:
                self.data.segments[i].update_dir(data_dir, data)
        else:
            utility.write_input_file(contents, new_levels, input_filename,
                output_dir)

        return input_filename, output_dir, data_dir

    def generate_workspace_extrap(self, theta, segment_indices=None):
        '''
        Similar to generate_workspace, except the test segments are updated
        rather than the data segments.
        '''
        contents = self.input_file_contents.copy()

        # Map theta to a new list of levels.
        new_levels = self.generate_levels(theta)

        # What extrapolation files need to be read?
        # If the user specifies the indices of the segments, then make sure
        # those are "include"d in the calculation and everything else is
        # excluded.
        t = Test('', contents=contents)
        if segment_indices is not None:
            for (i, test_segment) in enumerate(t.all_segments):
                test_segment.include = i in segment_indices
            t.write_segments(contents)

        # Write the updated contents to the input file and run.
        input_filename, output_dir = utility.random_output_dir_filename()
        utility.write_input_file(contents, new_levels, input_filename,
                                 output_dir)
        return input_filename, output_dir, t.get_output_files()



'''
The rows of levels in the .azr file are converted to list of strings. These
indices make it more convenient to access the desired parameter.
'''
J_INDEX = 0
PI_INDEX = 1
ENERGY_INDEX = 2
ENERGY_FIXED_INDEX = 3
CHANNEL_INDEX = 5
WIDTH_INDEX = 11
WIDTH_FIXED_INDEX = 10
SEPARATION_ENERGY_INDEX = 21
CHANNEL_RADIUS_INDEX = 27
OUTPUT_DIR_INDEX = 2
DATA_FILEPATH_INDEX = 11
LEVEL_INCLUDE_INDEX = 9


'''
Data-related constants.
'''
INCLUDE_INDEX = 0
IN_CHANNEL_INDEX = 1
OUT_CHANNEL_INDEX = 2
NORM_FACTOR_INDEX = 8
VARY_NORM_FACTOR_INDEX = 9
FILEPATH_INDEX = 11



'''
Utility functions stored here to keep other class definitions uncluttered.
'''

import string
import random
import os
from subprocess import Popen, PIPE
import numpy as np
from .level import Level
from .constants import *

def read_input_file(filename):
    '''
    Reads AZURE2 input file (.azr file) – purely for convenience.
    Takes a filename (str).
    Returns a list of strings.
    '''
    with open(filename, 'r') as f:
        contents = f.read().split('\n')
    return contents


def read_level_contents(infile):
    '''
    Reads rows between <levels> and </levels>.
    '''
    contents = read_input_file(infile)
    start = contents.index('<levels>')+1
    stop = contents.index('</levels>')
    return contents[start:stop]


def read_levels(infile):
    '''
    Packages the contents of the input file (infile, str) into instances of
    Level.
    Takes an input filename (str).
    Returns a list of Level instances.
    '''
    level_contents = read_level_contents(infile)

    levels = []
    sublevels = []
    for row in level_contents:
        if row != '':
            sublevels.append(Level(row))
        else:
            levels.append(sublevels)
            sublevels = []

    return levels


LETTERS = string.ascii_lowercase
NUMBERS = ''.join(map(str, range(10)))
CHARACTERS = LETTERS+NUMBERS


def random_string(length=8):
    return ''.join(random.choice(CHARACTERS) for i in range(length))


def random_output_dir_filename():
    s = 'mcazure_' + random_string()
    output_dir = 'output_' + s
    os.mkdir(output_dir)
    input_filename = s + '.azr'
    return input_filename, output_dir


def random_workspace(prepend=''):
    s = 'mcazure_' + random_string()
    output_dir = prepend + 'output_' + s
    data_dir = prepend + 'data_' + s
    os.mkdir(output_dir)
    os.mkdir(data_dir)
    input_filename = prepend + s + '.azr'
    return input_filename, output_dir, data_dir


def update_segmentsData_dir(contents0, data_dir):
    contents = contents0.copy()
    start = contents0.index('<segmentsData>')+1
    stop = contents0.index('</segmentsData>')

    for i in range(start, stop):
        row = contents[i].split()
        row[11] = row[11].replace('data', data_dir)
        contents[i] = ' '.join(row)
    
    return contents


def write_input_file(old_input_file_contents, new_levels, input_filename,
    output_dir, data_dir=None):
    '''
        Takes:
            * contents of an old .azr file (see read_input_file function)
            * list of new Levels
        Does:
            * replaces the level parameters of the old .azr files with the
              parameters of the new levels
            * generates a random filename
            * writes the new level parameters (along with everything else in the
              old .azr file) to the random filename
            * returns random filename
    '''
    start = old_input_file_contents.index('<levels>')+1
    stop = old_input_file_contents.index('</levels>')
    old_levels = old_input_file_contents[start:stop]
    nlines = len(old_levels)
    level_indices = [i for (i, line) in enumerate(old_levels) if line != '']
    nlevels = len(level_indices)
    blank_indices = [i for i in range(nlines) if i not in level_indices]
    assert (nlevels == len(new_levels)), '''
The number of levels passed in does not match the number of existing levels.'''

    # Replace the old level parameters with the new parameters.
    new_level_data = []
    j = 0
    for i in range(nlines):
        if i in blank_indices:
            new_level_data.append('')
        else:
            level = new_levels[j]
            nlevel = old_levels[i].split()
            nlevel[J_INDEX] = str(level.spin)
            nlevel[PI_INDEX] = str(level.parity)
            nlevel[ENERGY_INDEX] = str(level.energy)
            nlevel[WIDTH_INDEX] = str(level.width)
            nlevel[CHANNEL_RADIUS_INDEX] = str(level.channel_radius)
            new_level_data.append(str.join('  ', nlevel))
            j += 1

    # If the data directory is specified, then we'll update it.
    if data_dir is not None:
        old_input_file_contents = update_segmentsData_dir(old_input_file_contents, data_dir)

    # Write the new parameters to the same input file.
    with open(input_filename, 'w') as f:
        f.write(old_input_file_contents[0]+'\n')
        f.write(old_input_file_contents[1]+'\n')
        f.write(output_dir+'/\n')
        for row in old_input_file_contents[OUTPUT_DIR_INDEX+1:start]:
            f.write(row+'\n')
        for row in new_level_data:
            f.write(row+'\n')
        f.write('</levels>\n')
        for row in old_input_file_contents[stop+1:]:
            f.write(row+'\n')


def read_rwas_alt(output_dir):
    with open(output_dir + '/parameters.out', 'r') as f:
        pars = f.read().split('\n')
    rwas = []
    for row in pars:
        if row.find('g_int') >= 0:
            elements = row.split()
            ii = elements.index('g_int')
            rwas.append(float(elements[ii+2]))
    return rwas


def read_rwas_jpi(output_dir):
    with open(output_dir + '/parameters.out', 'r') as f:
        pars = f.read().split('\n')
    rwas = []
    for row in pars:
        if row.find('J =') == 0:
            Jpi = row.split()[2]
        if row.find('g_int') >= 0:
            elements = row.split()
            ii = elements.index('g_int')
            channel = int(elements[2])
            rwas.append([Jpi, channel, float(elements[ii+2])])
    return rwas


def read_ext_capture_file(filename):
    ext_capture_data = []
    with open(filename, 'r') as f:
        for line in f:
            i = line.find('(')
            j = line.find(',')
            k = line.find(')')
            x = float(line[i+1:j])
            y = float(line[j+1:k])
            ext_capture_data.append([x, y])
    return np.array(ext_capture_data)


def write_ext_capture_file(filename, data):
    '''
    data is expected to be 2-column matrix
    '''
    with open(filename, 'w') as f:
        for row in data:
            x, y = row
            f.write(f'({x:.5e},{y:.5e})\n')
        
    
def run_AZURE2(input_filename, choice=1, use_brune=False, ext_par_file='\n',
        ext_capture_file='\n', use_gsl=False, command='AZURE2'):
    cl_args = [command, input_filename, '--no-gui', '--no-readline']
    if use_brune:
        cl_args += ['--use-brune']
    if use_gsl:
        cl_args += ['--gsl-coul']
    p = Popen(cl_args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    options = str(choice) + '\n' + ext_par_file + ext_capture_file
    response = p.communicate(options.encode('utf-8'))
    return (response[0].decode('utf-8'), response[1].decode('utf-8'))
